@@ -1,6 +1,6 @@
 /* reading patches */
 
-/* $Id: pch.h,v 1.5 1997/04/07 01:07:00 eggert Exp $ */
+/* $Id: pch.h,v 1.7 1997/05/19 06:52:03 eggert Exp $ */
 
 LINENUM pch_end PARAMS ((void));
 LINENUM pch_first PARAMS ((void));
@@ -15,7 +15,8 @@ bool pch_write_line PARAMS ((LINENUM, FILE *));
 bool there_is_another_patch PARAMS ((void));
 char *pfetch PARAMS ((LINENUM));
 char pch_char PARAMS ((LINENUM));
-int another_hunk PARAMS ((enum diff));
+int another_hunk PARAMS ((enum diff, int));
+int pch_says_nonexistent PARAMS ((int));
 size_t pch_line_len PARAMS ((LINENUM));
 void do_ed_script PARAMS ((FILE *));
 void open_patch_file PARAMS ((char const *));
